#include <gba_interrupt.h>
#include <gba_systemcalls.h>
#include <gba_input.h>
#include <gba_video.h>

#include <math.h>

// The mode 2 framebuffer is at the start of VRAM
u16 * const framebuffer = (u16*)VRAM;

void plotPixel( u32 x, u32 y, u16 colour)
{
	framebuffer[y * SCREEN_WIDTH + x] = colour;
}

// starting position in centre of screen
#define START_X SCREEN_WIDTH/2
#define START_Y SCREEN_HEIGHT/2

double x = START_X, y = START_Y; 
// previous position so we only update when position changes
double lastX = START_X, lastY = START_Y;

// Initial direction is 90° or straight up the screen 
double pa = M_PI/2;
double pdx, pdy;

// start at rest
double velocity = 0;

// variables controlling our simple physics
// maximum speed we want to travel at - this is pixels per frame
static const double max_velocity = 2.5;
// how much we reduce speed each frame. This simulates friction of sorts
static const double deceleration = 0.01;
// how much speed our thruster adds each frame. Needs to be > deceleration or we won't move
static const double acceleration = 0.02;

// how far we want to turn each frame
// We have roughly 60 frames per second and 2π radians in a full circle
// dividing 2π by 120 means we take 120 frames or around 2 seconds for a full turn
static const double turnspeed = 2 * M_PI / 120;

int main(void) {
	// Set the video mode and enable background 2
	REG_DISPCNT = MODE_3 | BG2_ON;
	// Initialise libgba interrrupt dispatcher 
	irqInit();
	// Enable the vblank irq so we can use the system VBlankIntrWait.
	// libgba dispatcher will handle acknowledgement and setting flags for bios intrwait functions
	irqEnable(IRQ_VBLANK);

	// mode 2 is 16bit per pixel, 5 bits per component.
	// libgba provides RGB5 for readable colors 
	plotPixel(x, y, RGB5(31, 31, 0));

	while (1) {  // loop forever
		// libgba provides some basic button handling
		// scanKeys should be called once per loop
		scanKeys();
		// keysHeld tells us which buttons are currently pressed.
		// We also have keysDown and keysUp which tell us which buttons changed state in this loop
		int held = keysHeld();
		if (held & KEY_LEFT)
		{
			pa-=(turnspeed);
			if (pa < 0){ pa += 2*M_PI; } // wrap direction at one complete circle.
		}
		if (held & KEY_RIGHT)
		{
			pa+=(turnspeed);
			if (pa > 2*M_PI){ pa -= 2*M_PI; }
		}
		
		// A button increases velocity
		if (held & KEY_A) velocity += acceleration;
		// B button decreases it
		if (held & KEY_B) velocity -= acceleration;

		// we cap the velocity at something sensible 
		if (velocity > max_velocity) velocity = max_velocity;
		if (velocity < -max_velocity) velocity = -max_velocity;

		// Use trig to calculate x and y components of movement
		pdx = cos(pa) * velocity;
		pdy = sin(pa) * velocity;

		// Update our position
		x += pdx;
		y += pdy;

		// wrap our postion around the screen so we don't wander out of framebuffer
		if (x>SCREEN_WIDTH) x -= SCREEN_WIDTH;
		if (y>SCREEN_HEIGHT) y -= SCREEN_HEIGHT;
		if (x<0) x += SCREEN_WIDTH;
		if (y<0) y += SCREEN_HEIGHT;

		// If we're moving then apply our pseudo friction
		if (velocity > 0) velocity -= deceleration;
		if (velocity < 0) velocity += deceleration;

		// Now wait for the vblank period when framebuffer isn't being transferred to LCD
		VBlankIntrWait();

		// It's good practice to only update if something changed
		if (lastX != x || lastY != y) {
			// erase last position - make it black
			plotPixel(lastX, lastY, RGB5(0, 0, 0));
			// show current position - make it white 
			plotPixel(x, y, RGB5(31, 31, 31));
			lastX = x; lastY = y;
		}
	}
}
